           (*-------------------------------------------*
            |                October 2007               |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_law_basic_v4_1
imports CSP_F_law_basic Trace_v4_1
begin

(*****************************************************************

         1. 

 *****************************************************************)

(* --------------------------------------------------- *
       unfold only the first Sending and Receiving
 * --------------------------------------------------- *)

lemma cspF_first_Send_prefix:
  "a ! v -> P =F[M,M] a v -> P"
by (simp add: csp_prefix_ss_def)

lemma cspF_first_Rec_prefix:
  "a ? x:X -> Pf x =F[M,M] ? : (a ` X) -> (%x. (Pf ((inv a) x)))"
by (simp add: csp_prefix_ss_def)

lemma cspF_first_Int_pre_choice:
  "! :X -> Pf =F[M,M]  ! :X .. (%x. x -> (Pf x))"
by (simp add: csp_prefix_ss_def)

lemma cspF_first_Nondet_send_prefix:
  "a !? x:X -> Pf x =F[M,M] ! :(a ` X) -> (%x. (Pf ((inv a) x)))"
by (simp add: csp_prefix_ss_def)

lemmas cspF_first_prefix_ss =
       cspF_first_Send_prefix
       cspF_first_Rec_prefix
       cspF_first_Int_pre_choice
       cspF_first_Nondet_send_prefix

(* --------------------------------------------------- *
      Associativity of Sequential composition
 * --------------------------------------------------- *)

lemma cspT_Seq_compo_assoc:
  "(P ;; Q) ;; R =T[M,M] P ;; (Q ;; R)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)

 apply (rule)
 apply (simp add: in_traces)
 apply (elim disjE conjE exE)
  apply (force)

  apply (rule disjI2)
  apply (rule_tac x="sa" in exI)
  apply (insert noTick_or_last_Tick2)
  apply (drule_tac x="ta" in spec)
  apply (elim disjE conjE exE)

   apply (rule_tac x="ta" in exI)
   apply (simp)
   apply (rule disjI1)
   apply (rule_tac x="ta" in exI)
   apply (simp)

   apply (simp)
   apply (rule_tac x="tb" in exI)
   apply (simp)
   apply (rule disjI2)
   apply (rule_tac x="tb" in exI)
   apply (rule_tac x="<>" in exI)
   apply (simp)

   apply (subgoal_tac "noTick (s ^^ <Tick>)")
   apply (rotate_tac 3)
   apply (erule rem_asmE)
   apply (simp)
   apply (simp)

   apply (rule disjI2)
   apply (simp add: appt_decompo)
   apply (elim disjE conjE exE)

    apply (simp)
    apply (elim disjE conjE exE)
    apply (simp)
    apply (simp)
    apply (rule_tac x="sa" in exI)
    apply (rule_tac x="ta" in exI)
    apply (simp)
    apply (rule disjI2)
    apply (rule_tac x="<>" in exI)
    apply (rule_tac x="ta" in exI)
    apply (simp)

    apply (simp)
    apply (rotate_tac -2)
    apply (drule sym)
    apply (simp)
    apply (rotate_tac -2)
    apply (drule sym)
    apply (simp)

    apply (simp)
    apply (rule_tac x="sa" in exI)
    apply (rule_tac x="u ^^ ta" in exI)
    apply (simp add: appt_assoc)
    apply (rule disjI2)
    apply (rule_tac x="u" in exI)
    apply (rule_tac x="ta" in exI)
    apply (simp)

(* => *)
 apply (rule)
 apply (simp add: in_traces)
 apply (elim disjE conjE exE)

  apply (rule disjI1)
  apply (rule_tac x="t" in exI)
  apply (simp)
  apply (rule disjI1)
  apply (rule_tac x="s" in exI)
  apply (simp)

  apply (rule disjI1)
  apply (rule_tac x="s ^^ ta" in exI)
  apply (simp)
  apply (rule disjI2)
  apply (rule_tac x="s" in exI)
  apply (rule_tac x="ta" in exI)
  apply (simp)
  apply (rule memT_prefix_closed)
  apply (simp)
  apply (simp)

  apply (rule disjI2)
  apply (rule_tac x="s ^^ sa" in exI)
  apply (rule_tac x="tb" in exI)
  apply (simp add: appt_assoc)
  apply (rule disjI2)
  apply (rule_tac x="s" in exI)
  apply (rule_tac x="sa ^^ <Tick>" in exI)
  apply (simp)
done

lemma cspF_Seq_compo_assoc:
  "(P ;; Q) ;; R =F[M,M] P ;; (Q ;; R)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Seq_compo_assoc)
apply (rule order_antisym)

 apply (rule)
 apply (simp add: in_traces in_failures)
 apply (elim disjE conjE exE)
  apply (force)

  apply (rule disjI2)
  apply (simp)
  apply (rule_tac x="sa" in exI)
  apply (rule_tac x="t" in exI)
  apply (simp)

  apply (subgoal_tac "noTick (sa ^^ <Tick>)")
  apply (rotate_tac 3)
  apply (erule rem_asmE)
  apply (simp)
  apply (simp)

  apply (simp)
  apply (rule disjI2)
  apply (simp add: appt_decompo)
  apply (elim disjE conjE exE)

   apply (simp)
   apply (elim disjE conjE exE)
   apply (simp)
   apply (simp)
   apply (rule_tac x="sb" in exI)
   apply (rule_tac x="t" in exI)
   apply (simp)
   apply (rule disjI2)
   apply (rule_tac x="<>" in exI)
   apply (rule_tac x="t" in exI)
   apply (simp)

   apply (simp)
   apply (rotate_tac -2)
   apply (drule sym)
   apply (simp)
   apply (rotate_tac -2)
   apply (drule sym)
   apply (simp)

   apply (simp)
   apply (rule_tac x="sb" in exI)
   apply (rule_tac x="u ^^ t" in exI)
   apply (simp add: appt_assoc)
   apply (rule disjI2)
   apply (rule_tac x="u" in exI)
   apply (rule_tac x="t" in exI)
   apply (simp)

(* => *)
 apply (rule)
 apply (simp add: in_traces in_failures)
 apply (elim disjE conjE exE)

  apply (simp)

  apply (rule disjI1)
  apply (simp)
  apply (rule disjI2)
  apply (rule_tac x="sa" in exI)
  apply (rule_tac x="t" in exI)
  apply (simp)

  apply (rule disjI2)
  apply (rule_tac x="sa ^^ sb" in exI)
  apply (rule_tac x="ta" in exI)
  apply (simp add: appt_assoc)
  apply (rule disjI2)
  apply (rule_tac x="sa" in exI)
  apply (rule_tac x="sb ^^ <Tick>" in exI)
  apply (simp)
done

lemma cspF_Seq_compo_assoc_sym:
  "P ;; (Q ;; R) =F[M,M] (P ;; Q) ;; R"
apply (rule cspF_sym)
apply (simp add: cspF_Seq_compo_assoc)
done

end
