           (*-------------------------------------------*
            |           November 2007 for Isabelle 2005 |
            |           November 2008 for Isabelle 2008 |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_tactic_v4_1
imports CSP_F_tactic
        CSP_F_law_basic_v4_1  CSP_F_law_int_v4_1
        CSP_F_law_dist_v4_1   CSP_F_law_step_v4_1
        CSP_F_law_aux_v4_1    CSP_F_law_pipe_v4_1
begin

(*****************************************************************

         1. 

 *****************************************************************)

(* ----------------------------------------------------------- *

       new tactics (which will be added to the next version)

 * ----------------------------------------------------------- *)

lemmas cspF_all_dist_v4_1 = cspF_all_dist
                            cspF_Seq_compo_hide_dist
                            cspF_Interleave_hide_dist
                            cspF_Seq_compo_renaming_dist
                            cspF_Interleave_renaming_dist

                            cspF_dist_Alpha_Parallel
                            cspF_Dist_Alpha_Parallel

lemmas cspF_choice_IF_v4_1 = cspF_choice_IF cspF_Interleave_unit 
lemmas cspF_pre_step       = cspF_Alpha_Parallel_step

ML {*
 val CSPF_free_decompo_flag_v4_1     = thms "cspF_free_decompo_flag_v4_1" ;
 val CSPF_decompo_v4_1               = thms "cspF_decompo_v4_1" ;

 val CSPF_all_dist_v4_1              = thms "cspF_all_dist_v4_1" ;
 val CSPF_choice_IF_v4_1             = thms "cspF_choice_IF_v4_1" ;
 val CSPF_step_v4_1                  = thms "cspF_step_v4_1" ;

 val CSPF_choice_delay               = thms "cspF_choice_delay" ;          (* new *)
 val CSPF_Rep_int_choice_sepa        = thms "cspF_Rep_int_choice_sepa" ;   (* new *)
 val CSPF_Rep_int_choice_f_map       = thms "cspF_Rep_int_choice_f_map" ;  (* new *)

 val CSPF_first_prefix_ss            = thms "cspF_first_prefix_ss" ;
 val CSPF_prefix_Renaming_in_step    = thms "cspF_prefix_Renaming_in_step" ;
 val CSPF_prefix_Renaming_notin_step = thms "cspF_prefix_Renaming_notin_step" ;
 val CSPF_Renaming_fun_step          = thms "cspF_Renaming_step" ;

 val Split_if                        = thms "split_if" ;

 val CSPF_pre_step                   = thms "cspF_pre_step" ;             (* new *)

*}

(*-------------------------*
 |   deep simplification   |
 *-------------------------*)

ML {*
fun cspF_deep_simp_with_main_tac a n =
let val A = thms a
in
 CHANGED (
  EVERY
   [REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset () delsplits Split_if) n)),   (* no_use_asm, no split_if *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n,
                        resolve_tac A n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_decompo_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True delsplits Split_if)))])

end
*}

(*** deep_simp_with ***)

ML {*
fun cspF_deep_simp_with_left_tac a n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_deep_simp_with_main_tac a n])
*}

ML {*
fun cspF_deep_simp_with_right_tac a n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_deep_simp_with_main_tac a n])
*}

ML {*
    fun cspF_deep_simp_with_tac a n = 
      CHANGED (
       EVERY
        [TRY (cspF_deep_simp_with_left_tac a n),
         TRY (cspF_deep_simp_with_right_tac a n)])
*}


(*-------------------------*
 |     renaming (deep)     |
 *-------------------------*)

ML {*
fun cspF_ren_v4_1_main_tac n =
 CHANGED (
  EVERY
   [
    REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset () delsplits Split_if) n)),   (* no_use_asm, no split_if *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n,
                        resolve_tac CSPF_all_dist_v4_1 n,
                        resolve_tac CSPF_prefix_Renaming_in_step n,
                        resolve_tac CSPF_prefix_Renaming_notin_step n,
                        resolve_tac CSPF_Renaming_fun_step n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_decompo_v4_1 n ]),         (*  <--- deep *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

     (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True delsplits Split_if)))])
*}

(*** ren ***)

ML {*
fun cspF_ren_v4_1_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_ren_v4_1_main_tac n])
*}

ML {*
fun cspF_ren_v4_1_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_ren_v4_1_main_tac n])
*}

ML {*
    fun cspF_ren_v4_1_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_ren_v4_1_left_tac n),
         TRY (cspF_ren_v4_1_right_tac n)])
*}

(*-------------------------*
 |   renaming (deep) asm   |
 *-------------------------*)

ML {*
fun cspF_ren_asm_v4_1_main_tac n =
 CHANGED (
  EVERY
   [
    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset () delsplits Split_if) n)),   (* no split_if *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n,
                        resolve_tac CSPF_all_dist_v4_1 n,
                        resolve_tac CSPF_prefix_Renaming_in_step n,
                        resolve_tac CSPF_prefix_Renaming_notin_step n,
                        resolve_tac CSPF_Renaming_fun_step n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_decompo_v4_1 n ]),         (*  <--- deep *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

     (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True delsplits Split_if)))])
*}

(*** ren ***)

ML {*
fun cspF_ren_asm_v4_1_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_ren_asm_v4_1_main_tac n])
*}

ML {*
fun cspF_ren_asm_v4_1_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_ren_asm_v4_1_main_tac n])
*}

ML {*
    fun cspF_ren_asm_v4_1_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_ren_asm_v4_1_left_tac n),
         TRY (cspF_ren_asm_v4_1_right_tac n)])
*}

(*-------------------------*
 |       HSF ver.4.1       |
 *-------------------------*)

ML {*
fun cspF_hsf_v4_1_main_tac n =
 CHANGED (
  EVERY
   [
    REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset ()) n)),      (* no asm use *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n,
                        resolve_tac CSPF_all_dist_v4_1 n,
                        resolve_tac CSPF_prefix_Renaming_in_step n,
                        resolve_tac CSPF_prefix_Renaming_notin_step n,
                        resolve_tac CSPF_Renaming_fun_step n,
                        resolve_tac CSPF_first_prefix_ss n,
                        resolve_tac CSPF_pre_step n,
                        resolve_tac CSPF_SKIP_DIV_sort n,
                        resolve_tac CSPF_SKIP_DIV_resolve n,
                        resolve_tac CSPF_step_v4_1 n,
                        resolve_tac CSPF_choice_delay n,    (* new *)
                        resolve_tac CSPF_Rep_int_choice_sepa n,    (* new *)
                        resolve_tac CSPF_Rep_int_choice_f_map n    (* new *)
                       ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
*}

(*** hsf_v4_1 ***)

ML {*
fun cspF_hsf_v4_1_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_hsf_v4_1_main_tac n])
*}

ML {*
fun cspF_hsf_v4_1_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_hsf_v4_1_main_tac n])
*}
ML {*
    fun cspF_hsf_v4_1_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_hsf_v4_1_left_tac n),
         TRY (cspF_hsf_v4_1_right_tac n)])
*}

(*-------------------------*
 |     HSF ver.4.1  (asm)  |
 *-------------------------*)

ML {*
fun cspF_hsf_asm_v4_1_main_tac n =
 CHANGED (
  EVERY
   [
    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n,
                        resolve_tac CSPF_all_dist_v4_1 n,
                        resolve_tac CSPF_prefix_Renaming_in_step n,
                        resolve_tac CSPF_prefix_Renaming_notin_step n,
                        resolve_tac CSPF_Renaming_fun_step n,
                        resolve_tac CSPF_first_prefix_ss n,
                        resolve_tac CSPF_pre_step n,
                        resolve_tac CSPF_SKIP_DIV_sort n,
                        resolve_tac CSPF_SKIP_DIV_resolve n,
                        resolve_tac CSPF_step_v4_1 n,
                        resolve_tac CSPF_choice_delay n,    (* new *)
                        resolve_tac CSPF_Rep_int_choice_sepa n,    (* new *)
                        resolve_tac CSPF_Rep_int_choice_f_map n,    (* new *)
                        resolve_tac CSPF_unwind n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
*}

(*** hsf_asm_v4_1 ***)

ML {*
fun cspF_hsf_asm_v4_1_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_hsf_asm_v4_1_main_tac n])
*}

ML {*
fun cspF_hsf_asm_v4_1_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_hsf_asm_v4_1_main_tac n])
*}
ML {*
    fun cspF_hsf_asm_v4_1_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_hsf_asm_v4_1_left_tac n),
         TRY (cspF_hsf_asm_v4_1_right_tac n)])
*}

(*-------------------------*
 |       HSF ver.4.1       |
 *-------------------------*)

ML {*
fun cspF_hsf_unwind_v4_1_main_tac n =
 CHANGED (
  EVERY
   [
    REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset ()) n)),      (* no asm use *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n,
                        resolve_tac CSPF_all_dist_v4_1 n,
                        resolve_tac CSPF_prefix_Renaming_in_step n,
                        resolve_tac CSPF_prefix_Renaming_notin_step n,
                        resolve_tac CSPF_Renaming_fun_step n,
                        resolve_tac CSPF_first_prefix_ss n,
                        resolve_tac CSPF_pre_step n,
                        resolve_tac CSPF_SKIP_DIV_sort n,
                        resolve_tac CSPF_SKIP_DIV_resolve n,
                        resolve_tac CSPF_step_v4_1 n,
                        resolve_tac CSPF_choice_delay n,    (* new *)
                        resolve_tac CSPF_Rep_int_choice_sepa n,    (* new *)
                        resolve_tac CSPF_Rep_int_choice_f_map n,    (* new *)
                        resolve_tac CSPF_unwind n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
*}

(*** hsf_unwind_v4_1 ***)

ML {*
fun cspF_hsf_unwind_v4_1_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_hsf_unwind_v4_1_main_tac n])
*}

ML {*
fun cspF_hsf_unwind_v4_1_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_hsf_unwind_v4_1_main_tac n])
*}
ML {*
    fun cspF_hsf_unwind_v4_1_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_hsf_unwind_v4_1_left_tac n),
         TRY (cspF_hsf_unwind_v4_1_right_tac n)])
*}

(*-------------------------*
 |      simp ver.4.1       |
 *-------------------------*)

ML {*
fun cspF_simp_v4_1_main_tac n =
 CHANGED (
  EVERY
   [
    REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset ()) n)),      (* no asm use *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
*}

(*** simp_v4_1 ***)

ML {*
fun cspF_simp_v4_1_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_simp_v4_1_main_tac n])
*}

ML {*
fun cspF_simp_v4_1_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_simp_v4_1_main_tac n])
*}

ML {*
    fun cspF_simp_v4_1_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_simp_v4_1_left_tac n),
         TRY (cspF_simp_v4_1_right_tac n)])
*}


(*-------------------------*
 |    simp_with ver.4.1    |
 *-------------------------*)

ML {*
fun cspF_simp_with_v4_1_main_tac a n =
let val A = thms a
in
 CHANGED (
  EVERY
   [REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset ()) n)),      (* no asm use *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n,
                        resolve_tac A n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
end
*}

(*** simp_with_v4_1 ***)

ML {*
fun cspF_simp_with_v4_1_left_tac a n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_simp_with_v4_1_main_tac a n])
*}

ML {*
fun cspF_simp_with_v4_1_right_tac a n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_simp_with_v4_1_main_tac a n])
*}
ML {*
    fun cspF_simp_with_v4_1_tac a n = 
      CHANGED (
       EVERY
        [TRY (cspF_simp_with_v4_1_left_tac a n),
         TRY (cspF_simp_with_v4_1_right_tac a n)])
*}

(*-----------------------------*
 |   refine_with left ver.4.1  |
 *-----------------------------*)

(* left *)

ML {*
fun cspF_refine_with_v4_1_left_main_tac a n =
let val A = thms a
in
 CHANGED (
  EVERY
   [REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset ()) n)),      (* no asm use *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac A n]),

       (EVERY [ resolve_tac CSPF_tr_flag_left n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),


       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
end
*}

(*** refine_with ***)

ML {*
fun cspF_refine_with_v4_1_left_tac a n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_tr_flag_left n,
    cspF_refine_with_v4_1_left_main_tac a n])
*}

(*-----------------------------*
 |  refine_with right ver.4.1  |
 *-----------------------------*)

ML {*
fun cspF_refine_with_v4_1_right_main_tac a n =
let val A = thms a
in
 CHANGED (
  EVERY
   [REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset ()) n)),     (* no asm use *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac A n ]),

       (EVERY [ resolve_tac CSPF_tr_flag_right n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
end
*}

(*** refine_with ***)

ML {*
fun cspF_refine_with_v4_1_right_tac a n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_tr_flag_right n,
    cspF_refine_with_v4_1_right_main_tac a n])
*}

(*----------------------------*
 |   refine_asm left ver.4.1  |
 *----------------------------*)

(* left *)

ML {*
fun cspF_refine_asm_v4_1_left_main_tac n =
 CHANGED (
  EVERY
   [REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ resolve_tac CSPF_tr_flag_left n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
*}

(*** refine_asm ***)

ML {*
fun cspF_refine_asm_v4_1_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_tr_flag_left n,
    cspF_refine_asm_v4_1_left_main_tac n])
*}

(*----------------------------*
 |  refine_asm right ver.4.1  |
 *----------------------------*)

ML {*
fun cspF_refine_asm_v4_1_right_main_tac n =
 CHANGED (
  EVERY
   [REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ resolve_tac CSPF_tr_flag_right n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
*}

(*** refine_asm ***)

ML {*
fun cspF_refine_asm_v4_1_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_tr_flag_right n,
    cspF_refine_asm_v4_1_right_main_tac n])
*}

(*-------------------------*
 |     Unwind ver.4.1      |
 *-------------------------*)

ML {*
fun cspF_unwind_v4_1_main_tac n =
 CHANGED (
  EVERY
   [
    REPEAT (
     FIRST 
      [(CHANGED (full_simp_tac (simpset ()) n)),      (* no asm use *)

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_IF_v4_1 n,
                        resolve_tac CSPF_unwind n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag_v4_1 n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                TRY (resolve_tac OFF_Not_Rewrite_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_All_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_All_Flag_True)))])
*}

(*** unwind_v4_1 ***)

ML {*
fun cspF_unwind_v4_1_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,
    cspF_unwind_v4_1_main_tac n])
*}

ML {*
fun cspF_unwind_v4_1_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,
    cspF_unwind_v4_1_main_tac n])
*}
ML {*
    fun cspF_unwind_v4_1_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_unwind_v4_1_left_tac n),
         TRY (cspF_unwind_v4_1_right_tac n)])
*}

end
